package xyz.matj1.bp.experiments;

import mcdr.objects.utils.RankedSortedCollectionDistHashcode;
import mcdr.test.utils.ObjectMgmt;
import messif.objects.LocalAbstractObject;
import messif.objects.util.RankedSortedCollection;
import messif.operations.AnswerType;
import messif.operations.query.KNNQueryOperation;
import xyz.matj1.bp.Config;
import xyz.matj1.bp.StretchedSequences;
import xyz.matj1.bp.enums.MocapType;
import xyz.matj1.bp.enums.SpeedMultiple;
import xyz.matj1.bp.exceptions.CacheIOException;
import xyz.matj1.bp.types.MyMap;

import java.io.File;
import java.io.IOException;
import java.util.Collection;
import java.util.Collections;

import static xyz.matj1.bp.Utils.loadData;
import static xyz.matj1.bp.Utils.readCache;
import static xyz.matj1.bp.Utils.writeCache;
import static xyz.matj1.bp.enums.SpeedMultiple._1;

public class KnnSearchTools {
	public static MyMap<LocalAbstractObject, Float> knnSearch(
		LocalAbstractObject queryObject,
		Collection<LocalAbstractObject> searchSpace,
		Integer k
	) {
		return knnSearch(
			queryObject,
			searchSpace,
			k,
			false,
			AnswerType.ORIGINAL_OBJECTS,
			new RankedSortedCollectionDistHashcode()
		);
	}
	
	public static MyMap<LocalAbstractObject, Float> knnSearch(
		LocalAbstractObject queryObject,
		Collection<LocalAbstractObject> searchSpace,
		Integer k,
		boolean storeMetaDistances,
		AnswerType answerType,
		RankedSortedCollection answerCollection
	) {
		var operation = new KNNQueryOperation(
			queryObject,
			k,
			storeMetaDistances,
			answerType,
			answerCollection
		);

		searchSpace.forEach(
			operation::addToAnswer
		);

		var result = new MyMap<LocalAbstractObject, Float>();
		operation.forEach(a -> result.put((LocalAbstractObject)a.getObject(), a.getDistance()));

		return result;
	}

	/**
	 * For each mocap type, for each speed multiple, this returns a map from each sequence
	 * to the list of sequences found by kNN search around the sequence
	 * with k equal to the number of other sequences in the category of the sequence.
	 *
	 * This caches results for faster retrieval after the first call, unless the caches are deleted or invalid.
	 *
	 * @return a nested map for each mocap type for each speed multiple, for each sequence, a list of the sequences nighest to the sequence
	 */
	public static MyMap<MocapType, MyMap<SpeedMultiple, MyMap<LocalAbstractObject, MyMap<LocalAbstractObject, Float>>>> stretchKnnResults() throws IOException {
		var results = new MyMap<MocapType, MyMap<SpeedMultiple, MyMap<LocalAbstractObject, MyMap<LocalAbstractObject, Float>>>>();

		for(var type : MocapType.values()) {
			var multiple_Results = new MyMap<SpeedMultiple, MyMap<LocalAbstractObject, MyMap<LocalAbstractObject, Float>>>();

			for(var multiple : SpeedMultiple.values()) {
				multiple_Results.put(multiple, stretchKnnResults(type, multiple));
			}

			results.put(type, multiple_Results);
		}

		return results;

	}

	public static MyMap<LocalAbstractObject, MyMap<LocalAbstractObject, Float>> stretchKnnResults(
		MocapType type,
		SpeedMultiple multiple
	) throws IOException {
		var cacheDir = new File(Config.cacheDir(), "stretchKnnResults");
		cacheDir.mkdirs();
		var cachePath = new File(cacheDir, String.format("%s – %s×", type.noun(), multiple.asString()));

		try {
			return readCache(cachePath, MyMap.class);
		} catch (CacheIOException _e) {
			var stretchedSequenceMgmt = loadData(
				StretchedSequences.dataFiles(type).get(multiple),
				type.dataClass()
			);
			var baseSequenceMgmt = loadData(
				StretchedSequences.dataFiles(type).get(_1),
				type.dataClass()
			);

			MyMap<LocalAbstractObject, MyMap<LocalAbstractObject, Float>> results = stretchKnnResultsCompute(
				stretchedSequenceMgmt,
				baseSequenceMgmt
			);

			try {
				writeCache(cachePath, results);
			} catch (CacheIOException e) {
				System.err.printf("Writing cache to %s failed.%n", cachePath);
				e.printStackTrace();
			}

			return results;
		}
	}

	private static MyMap<LocalAbstractObject, MyMap<LocalAbstractObject, Float>> stretchKnnResultsCompute(
		ObjectMgmt stretchedSequenceMgmt,
		ObjectMgmt baseSequenceMgmt
	) {
		var baseObjects = Collections.unmodifiableSet(baseSequenceMgmt.getObjects());
		var baseCategory_Objects = Collections.unmodifiableMap(baseSequenceMgmt.getCategoryObjects());
		var stretchedObjects = Collections.unmodifiableSet(stretchedSequenceMgmt.getObjects());
		var results = new MyMap<LocalAbstractObject, MyMap<LocalAbstractObject, Float>>();

		baseCategory_Objects
			.entrySet()
			.parallelStream()
			.forEach(entry -> {
				var sequencesInCategory = entry.getValue();
				var categorySize = sequencesInCategory.size();

				sequencesInCategory.parallelStream().forEach(sequence -> {
					var stretchedObject = stretchedObjects
						.stream().filter(
							StretchedSequences.uriMatches(sequence)
						).findAny().orElseThrow();

					var knnSequence_Distance = KnnSearchTools.knnSearch(
						stretchedObject,
						baseObjects.stream()
							.filter(StretchedSequences.uriMatches(stretchedObject).negate())
							.toList(),
						Math.max(20, categorySize - 1)
					);
					
					results.put(sequence, knnSequence_Distance);
				});
			});
		return results;
	}
}
